import json
import os
from typing import Dict, List, Set
import sharepp
from investment_rebalancer.model.errors import ConfigurationException
from investment_rebalancer.model.asset.etf import ETF
from investment_rebalancer.model.classification.category import Category
from investment_rebalancer.model.classification.classification import Classification

etfs: Dict[str, ETF] = {}
classifications: List[Classification] = []


def parse(config_path: str):
    if not os.path.isfile(config_path):
        raise ConfigurationException(
            f"The given configuration file does not exist: {config_path}"
        )

    with open(config_path) as configuration_file:
        config = json.load(configuration_file)

    # Read etf configs
    etf_config = config["etf"]
    for etf in etf_config:
        try:
            name = etf_config[etf]["name"]
            print(f"Getting current price for {name}")
            current_price = sharepp.get_etf_price(etf)
            # current_price = 5.0
            etfs[etf] = ETF(
                isin=etf,
                name=name,
                enabled=etf_config[etf]["enabled"],
                quantity=etf_config[etf]["quantity"],
                current_price=current_price,
                ter=etf_config[etf]["ter"],
            )
        except KeyError as e:
            raise ConfigurationException(f"Key {str(e)} missing in ETF {etf}!")

    if not etfs:
        raise ConfigurationException("No ETFs configured!")

    # Read classification configs
    classification_config = config["classifications"]
    for classification in classification_config:
        try:
            classifications.append(
                parse_classification(
                    classification, classification_config[classification]
                )
            )
        except KeyError as e:
            raise ConfigurationException(
                f"Key {str(e)} missing in classification {classification}!"
            )

    if not classifications:
        raise ConfigurationException("No classifications configured!")


def parse_classification(name: str, categories_config) -> Classification:
    categories: Set[Category] = []
    for category_config in categories_config:
        assets: List[ETF] = []
        for asset_config in categories_config[category_config]["assets"]:
            assets.append(etfs[asset_config])
        categories.append(
            Category(
                category_config,
                categories_config[category_config]["target_allocation"],
                assets,
            )
        )
    return Classification(name, categories)


def get_investable_etfs() -> Set[ETF]:
    investable_etfs: Set[ETF] = []
    for isin in etfs:
        investible = True
        for classification in classifications:
            if not classification.is_investible(etfs[isin]):
                investible = False
                break
        if investible:
            investable_etfs.append(etfs[isin])
    return investable_etfs


def get_all_categories() -> List[Category]:
    categories: List[Category] = []
    for classification in classifications:
        for category in classification.categories:
            if category.to_invest > 0.0:
                categories.append(category)
    categories.sort(key=lambda x: x.to_invest)
    return categories



from typing import List, Set
from investment_rebalancer.model.asset.etf import ETF


class Category:
    def __init__(self, name: str, target_allocation: float, etfs: List[ETF]) -> None:
        self.name = name
        self.target_allocation = target_allocation
        self.etfs = etfs
        self.etfs.sort(key=lambda x: x.ter)
        self.current_allocation: float = 0.0
        self.delta_value: float = 0.0
        self.to_invest: float = 0.0

    @property
    def current_value(self) -> float:
        value = 0.0
        for etf in self.etfs:
            value += etf.current_value
        return value

    def invest(self, investment: float, investable_etfs: Set[ETF]) -> float:
        for etf in self.etfs:
            if etf in investable_etfs and etf.enabled:
                if self.to_invest < investment:
                    etf.investment += self.to_invest
                    new_investment = investment - self.to_invest
                    self.to_invest = 0.0
                    return new_investment
                else:
                    etf.investment += investment
                    self.to_invest -= investment
                    return 0.0
        return investment

    def __eq__(self, other: object) -> bool:
        """
        Checks whether or not an object is equal to this instance.
        Two categories are equal if they have the same name.
        :param other: the other object
        :return: true if the other object is equal, otherwise false
        """
        if other == self:
            return True
        if not isinstance(other, Category):
            return False
        return other.name == self.name



import argparse
import os
from pathlib import Path

from prettytable import PrettyTable
from prompt_toolkit import prompt

from investment_rebalancer.controller.number_validator import NumberValidator
from investment_rebalancer.model import configuration
from investment_rebalancer.model.errors import ConfigurationException


def main():
    parser = argparse.ArgumentParser(
        description="Calculate investments for rebalancing."
    )
    parser.add_argument(
        "-c",
        "--config",
        default=os.path.join(Path.home(), ".investment-rebalancer/config.json"),
    )
    args = parser.parse_args()

    try:
        configuration.parse(args.config)
    except ConfigurationException as e:
        print(f"The configuration could not be read: {str(e)}")
        return

    total_value = 0.0
    for etf in configuration.etfs:
        total_value += configuration.etfs[etf].current_value

    print(f"Total investment value: {total_value:.2f}€\n")

    for classification in configuration.classifications:
        print(f"{classification.name}:")
        table = PrettyTable(["Category", "Value", "Allocation", "Target"])
        for category in classification.categories:
            current_allocation = category.current_value / total_value * 100
            table.add_row(
                [
                    category.name,
                    f"{category.current_value:.2f}€",
                    f"{current_allocation:.2f}%",
                    f"{category.target_allocation:.2f}%",
                ]
            )

        table.align = "l"
        print(table.get_string(sortby="Allocation", reversesort=True))

    investment_value = float(
        prompt("\nHow much money do you want to invest? ", validator=NumberValidator())
    )

    for classification in configuration.classifications:
        classification.calculate_target_values(investment_value)

    # Print all categories with a delta.
    print("\nCategories to invest in:")
    for classification in configuration.classifications:
        for category in classification.categories:
            if category.to_invest > 0.0:
                print(f"{category.name}: {category.to_invest:.2f}€")

    investable_categories = configuration.get_all_categories()
    while len(investable_categories) > 0 and round(investment_value, 2) > 0.0:
        investment_value = investable_categories[0].invest(
            investment_value, configuration.get_investable_etfs()
        )
        investable_categories = configuration.get_all_categories()

    # Print all ETF to invest in.
    print("\nETFs to invest in:")
    for etf in configuration.etfs.values():
        if etf.investment > 0.0:
            print(f"{etf.name}: {round(etf.investment)}€")


if __name__ == "__main__":
    main()



from typing import List

from investment_rebalancer.model.classification.category import Category
from investment_rebalancer.model.asset.etf import ETF


class Classification:
    def __init__(self, name: str, categories: List[Category]) -> None:
        self.name = name
        self.categories = categories

    @property
    def current_value(self) -> float:
        value = 0.0
        for category in self.categories:
            value += category.current_value
        return value

    def __eq__(self, other: object) -> bool:
        """
        Checks whether or not an object is equal to this instance.
        Two classifications are equal if they have the same name.
        :param other: the other object
        :return: true if the other object is equal, otherwise false
        """
        if other == self:
            return True
        if not isinstance(other, Classification):
            return False
        return super().__eq__(other)

    def calculate_target_values(self, investment_value: float):
        target_value = self.current_value + investment_value
        for category in self.categories:
            category_target_value = target_value * (category.target_allocation / 100)
            category.delta_value = category_target_value - category.current_value
            if category.delta_value > investment_value:
                category.to_invest += investment_value
                investment_value = 0
            elif category.delta_value > 0:
                category.to_invest += category.delta_value
                investment_value -= category.delta_value

    def is_investible(self, etf: ETF) -> bool:
        for category in self.categories:
            if etf in category.etfs and category.to_invest <= 0.0:
                return False
        return True



import sharepp
from pydantic import BaseModel, Field, computed_field, field_validator


class ETF(BaseModel):
    isin: str
    name: str
    enabled: bool = Field(default=False)
    quantity: float = Field(default=0.0)
    current_price: float
    ter: float
    investment: float = Field(default=0.0)

    @computed_field
    @property
    def current_value(self) -> float:
        return self.quantity * self.current_price

    @field_validator("isin")
    @classmethod
    def name_must_contain_space(cls, isin: str) -> str:
        if not sharepp.is_isin(isin):
            raise ValueError(f"{isin} is not a valid ISIN!")
        return isin

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ETF):
            return False
        return other.isin == self.isin



from prompt_toolkit.validation import Validator, ValidationError


class NumberValidator(Validator):
    def validate(self, document):
        text = document.text

        if text and not text.isdigit():
            raise ValidationError(message="This input contains non-numeric characters")
